"""
Created by Tetsu Haruyama
"""

import numpy as np
from scipy.stats import truncnorm
import statsmodels.api as sm
from py4etrics.base_for_models import GenericLikelihoodModel_TobitTruncreg

class Truncreg(GenericLikelihoodModel_TobitTruncreg):
    """
    Method 1:
    Truncreg(endog, exog, left=<-np.inf>, right=<np.inf>).fit()
    endog = dependent variable
    exog = independent variable (add constant if needed)
    left = the threshold value for left-truncation (default:-np.inf)
    right = the threshold value for right-truncation (default:np.inf)

    Method 2:
    formula = 'y ~ 1 + x'
    Truncreg(formula, left=<-np.inf>, right=<np.inf>, data=<DATA>).fit()

    Note:
    Left-truncated Regression if 'left' only is set.
    Right-truncated Regression if 'right' only is set.
    Left- and Right-truncated Regression if 'left' and 'right' both are set.

    """

    def __init__(self, endog, exog, left=None, right=None, **kwds):
        super(Truncreg, self).__init__(endog, exog, **kwds)

        if left == None:
            left = -np.inf
        self.left = left

        if right == None:
            right = np.inf
        self.right = right

    def loglikeobs(self, params):
        s = params[-1]
        beta = params[:-1]

        def _truncreg(y,x,left,right,beta,s):
            Xb = np.dot(x, beta)
            _l = (left - Xb)/np.exp(s)
            _r = (right - Xb)/np.exp(s)
            return truncnorm.logpdf(y,a=_l,b=_r,loc=Xb,scale=np.exp(s))

        return _truncreg(self.endog, self.exog,
                         self.left, self.right, beta, s)


    def fit(self, cov_type='nonrobust', start_params=None, maxiter=10000, maxfun=10000, **kwds):
        # add sigma for summary
        if 'Log(Sigma)' not in self.exog_names:
            self.exog_names.append('Log(Sigma)')
        else:
            pass
        # initial guess
        res_ols = sm.OLS(self.endog, self.exog).fit()
        params_ols = res_ols.params
        sigma_ols = np.log(np.std(res_ols.resid))
        if start_params == None:
            start_params = np.append(params_ols, sigma_ols)

        return super(Truncreg, self).fit(cov_type=cov_type, start_params=start_params,
                                     maxiter=maxiter, maxfun=maxfun, **kwds)

# EOF



"""
Created by Tetsu Haruyama
"""

import warnings
import numpy as np
from scipy.stats import norm
import statsmodels.api as sm
from py4etrics.base_for_models import GenericLikelihoodModel_TobitTruncreg

class Tobit(GenericLikelihoodModel_TobitTruncreg):
    """
    Create a Seires or array (to be used for "<censor>" below)
           -1: left-censored
            0: uncensored
            1: right-censored

    Method 1:
    Tobit(endog, exog, cens=<censor>, left=<0>, right=<0>).fit()
    endog = dependent variable
    exog = independent variable (add constant if needed)
    cens = see above
    left = the threshold value for left-censoring (default:0)
        　　(This becomes effective only if -1 is set in <censor>)
    right = the threshold value for right-censoring (default:0)
        　　(This becomes effective only if 1 is set in <censor>)
          
    Method 2:
    formula = 'y ~ 1 + x'
    Tobit.from_formula(formula, cens=<censor>, left=<0>, right=<0>, data=<DATA>).fit()
  
  """

    def __init__(self, endog, exog, cens, left=None, right=None, **kwds):
        super(Tobit, self).__init__(endog, exog, **kwds)
        self.cens = cens
        if left is None:
            left = 0
        self.left = left
        if right is None:
            right = 0
        self.right = right

    def loglikeobs(self, params): # see _tobit()
#     def loglike(self, params): # see _tobit()

        s = params[-1]
        beta = params[:-1]

        def _tobit(y, x, z, left, right, beta, s):
            if ( ~np.isin(z,[-1,0,1]) ).any():
                warnings.warn('\n\n***************************************************\n\n'+
                              'There are values other than [-1,0,1] in a column for cens\n\n'+
                             '***************************************************\n')

            # indicators
            left_on  = np.where(z==-1, 1, 0)
            mid_on   = np.where(z== 0, 1, 0)
            right_on = np.where(z== 1, 1, 0)

            Xb = np.dot(x, beta)

            # Eqn 5 of https://cran.r-project.org/web/packages/censReg/vignettes/censReg.pdf
            # scale=np.exp(s)
            left_mle  = left_on *  norm.logcdf( (left-Xb)  / np.exp(s) )
            mid_mle   = mid_on * ( norm.logpdf( (y-Xb)     / np.exp(s) ) - s )
            right_mle = right_on * norm.logcdf( (Xb-right) / np.exp(s) )
            
            return left_mle + mid_mle + right_mle  #  loglikeobs()                        
#             return (left_mle+mid_mle+right_mle).sum()  #  loglike()

        return _tobit(self.endog, self.exog, self.cens, self.left, self.right, beta, s)

    def fit(self, cov_type='nonrobust', start_params=None, maxiter=10000, maxfun=10000, **kwds):
        
        # add sigma for summary
        if 'Log(Sigma)' not in self.exog_names:  # ML Case 1 is used
            self.exog_names.append('Log(Sigma)')
        else:
            pass
        
        # initial guess
        res = sm.OLS(self.endog, self.exog).fit()
        ols_params = res.params
        ols_sigma = np.log(np.std(res.resid))
        
        # option
        if start_params == None:
            start_params = np.append(ols_params, ols_sigma)

        res = super(Tobit, self).fit(cov_type=cov_type, start_params=start_params,
                                     maxiter=maxiter, maxfun=maxfun, **kwds)
        return res



'''
Specification test for heteroscedasticity in Logit/Probit models

The tests are based on
(1) Wooldridge 2010, section 15.5.3
(2) https://www.statalist.org/forums/forum/general-stata-discussion/general/1292180-test-for-heteroskedasticity-in-logit-probit-models

Check consistency with Stata using the following

----- Stata code starts -----

// https://www.statalist.org/forums/forum/general-stata-discussion/general/1292180-test-for-heteroskedasticity-in-logit-probit-models
// https://www.stata.com/manuals13/p_predict.pdf
// https://www.stata.com/manuals13/rpredict.pdf

import delimited "/Users/Tetsu/Documents/My_Simulation/Python/projects/wooldridge/raw_data/data_csv/mroz.csv"

probit inlf nwifeinc educ exper expersq age kidslt6 kidsge6

predict xbhat, xb
// the following give s the same resutl
// predict xbhat, index

// 'c' for continuous variables for interaction term
// '#' multiplication for interaction terms
probit inlf nwifeinc educ exper expersq age kidslt6 kidsge6 c.xbhat#c.nwifeinc c.xbhat#c.educ c.xbhat#c.exper c.xbhat#c.expersq c.xbhat#c.age c.xbhat#c.kidslt6 c.xbhat#c.kidsge6

test c.xbhat#c.nwifeinc c.xbhat#c.educ c.xbhat#c.exper c.xbhat#c.expersq c.xbhat#c.age c.xbhat#c.kidslt6 c.xbhat#c.kidsge6

// for reference ---------------------------------------------------------------
hetprobit inlf nwifeinc educ exper expersq age kidslt6 kidsge6, het(nwifeinc educ exper expersq age kidslt6 kidsge6)

----- Stata code ends -----

Created by Tetsugen Haruyama
'''


import numpy as np
from scipy import stats
import pandas as pd
import statsmodels.api as sm
from statsmodels.formula.api import logit, probit


def het_test_logit(results):
    """
    Wald Test for Logit
    -------------------
    H0: homoscedasticity
    HA: heteroscedasticity

    Parameters
    ----------
    results : Logit results instance


    Returns
    -------
    Wald test statistic
    p-value
    Degree of Freedom
        The number of restrictions, which are equivalent to the number of explanatory variables, excluding a constant term

    References
    ----------
    The test is based on
    (1) Wooldridge 2010, section 15.5.3
    (2) https://www.statalist.org/forums/forum/general-stata-discussion/general/1292180-test-for-heteroskedasticity-in-logit-probit-models
    """

    yhat = results.predict(linear=True)  # original fitted values
    exog_var = results.model.exog  # original exog
    exog_df = pd.DataFrame(exog_var)  # convert to DataFrame

    try:  # drop a column of a constant if any
        tt = exog_df.nunique()
        idx_1 = list(tt).index(1.0)
        exog_df = exog_df.drop(idx_1, axis=1)
    except ValueError:
        pass

    num_para = exog_df.shape[1] # no of non-constant parameters

    # X = np.exp(yhat).reshape(len(yhat),1) * exog_df.values
    X = yhat.reshape(len(yhat),1) * exog_df.values

    endog = results.model.endog
    exog = np.column_stack((results.model.exog, X))
    res_test = sm.Logit(endog,exog).fit(disp=False)

    A = np.identity(len(res_test.params))
    A = A[-num_para:,:]
    s = res_test.wald_test(A)
    return print('H0: homoscedasticity\nHA: heteroscedasticity\n',
                '\nWald test:', "%#2.3f" % s.statistic[0][0],
                '\np-value:', "%#7.3f" % s.pvalue, '\ndf freedom:', "%#3.0f" % s.df_denom)





def het_test_probit(results):
    """
    Wald検定 for Probit
    ------------------
    H0: homoscedasticity
    HA: heteroscedasticity

    Parameters
    ----------
    results : Logit results instance

    Returns
    -------
    Wald test statistic
    p-value
    Degree of Freedom
        The number of restrictions, which are equivalent to the number of explanatory variables, excluding a constant term

    References
    ----------
    The test is based on
    (1) Wooldridge 2010, section 15.5.3
    (2) https://www.statalist.org/forums/forum/general-stata-discussion/general/1292180-test-for-heteroskedasticity-in-logit-probit-models
    """

    yhat = results.predict(linear=True)  # original fitted values
    exog_var = results.model.exog  # original exog
    exog_df = pd.DataFrame(exog_var)  # convert to DataFrame

    try:  # drop a column of a constant if any
        tt = exog_df.nunique()
        idx_1 = list(tt).index(1.0)
        exog_df = exog_df.drop(idx_1, axis=1)
    except ValueError:
        pass

    num_para = exog_df.shape[1] # no of non-constant parameters

    # X = np.exp(yhat).reshape(len(yhat),1) * exog_df.values
    X = yhat.reshape(len(yhat),1) * exog_df.values

    endog = results.model.endog
    exog = np.column_stack((results.model.exog, X))
    res_test = sm.Probit(endog,exog).fit(disp=False)

    A = np.identity(len(res_test.params))
    A = A[-num_para:,:]
    s = res_test.wald_test(A)
    return print('H0: homoscedasticity\nHA: heteroscedasticity\n',
                '\nWald test:', "%#2.3f" % s.statistic[0][0],
                '\np-value:', "%#7.3f" % s.pvalue, '\ndf freedom:', "%#3.0f" % s.df_denom)



from setuptools import setup, find_packages

setup(name='py4etrics',
      version='0.1.7',
      author='Tetsu HARUYAMA',
      author_email='haruyama@econ.kobe-u.ac.jp',
      packages=find_packages(),
      package_dir={'py4etrics': './py4etrics'},
      url='https://github.com/Py4Etrics/py4etrics',
      license='MIT',
      description='A package for py4etrics.github.io',
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      classifiers=['Intended Audience :: Education',
                   'Intended Audience :: Science/Research',
                   'License :: OSI Approved :: MIT License',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python',
                   ],
      keywords=['Truncated Regression', 'Tobit Model', 'Heckit Model']
      )



